//
//  histogram.js
//  Library
//
//  Created by Jonathon Mah on 2013-03-18.
//  Copyright (c) 2013 Delicious Monster Software. All rights reserved.
//


function Histogram(chartDiv, svg, width, height) {
    this.chartDiv = chartDiv;
    chartDiv.classed("histogram", true);
    this.svg = svg;
    this.width = width;
    this.height = height;

    this.dataFunction = function() { return []; };
    this.barColor = "steelblue";
    this.xBins = 20;
    this.avoidOutliers = false;

    this.axisSize = [32, 18]; // x,y space to show axis and labels
    this.axisPadding = [9, 5]; // extra space to avoid cutting off labels

    this.yScale = d3.scale.linear();

    this.xAxis = d3.svg.axis()
        .ticks(6)
        .orient("bottom");

    this.yAxis = d3.svg.axis()
        .scale(this.yScale)
        .orient("left")
        .ticks(5);

    this.histogram = d3.layout.histogram()
        .frequency(true);
}


Histogram.prototype.update = function() {
    var self = this;
    var values = this.dataFunction().sort(function(a,b) { return a - b; });
    if (!values.length) {
        this.svg
            .selectAll("g.bar")
            .remove();
        this.svg
            .select("g.x.axis")
            .remove();
        this.svg
            .select("g.y.axis")
            .remove();
        return;
    }


    // Pick a maximum value for the chart, attempting to avoid an outlier bunching up the bins.
    var min = values[0];
    var max = values[values.length - 1];
    var newXScale = d3.scale.linear()
        .range([this.axisSize[0], this.width - this.axisPadding[0]])
        .domain([min, max]).nice();

    if (this.avoidOutliers) {
        var q98 = d3.quantile(values, 0.98);

        var q100ScaleMax = d3.max(newXScale.domain());
        newXScale.domain([min, q98]).nice();
        var q98ScaleMax = d3.max(newXScale.domain());

        if (q98ScaleMax < q100ScaleMax)
            for (var i = 0; i < values.length; i++)
                if (values[i] > q98ScaleMax)
                    values[i] = q98ScaleMax;
    }

    if (this.xScale === undefined)
        this.xScale = newXScale;
    else if (values.length == 0)
        newXScale = this.xScale;

    this.histogram
        .range(newXScale.domain())
        .bins(this.xBins);

    var barMargin = 1;
    var data = this.histogram(values);

    this.yScale
        .domain([0, d3.max(data, function(d) { return d.y; })])
        .rangeRound([this.height - this.axisSize[1], this.axisPadding[1]]);

    var bars =
        this.svg
        .selectAll("g.bar")
        .data(data, function(d) { return "" + d.x + "," + d.dx; });


    // insert
    var enteredBars =
        bars
        .enter()
        .append("g")
        .classed("bar", true);

    enteredBars
        .append("rect")
        .attr("fill", this.barColor)
        .call(positionBar)
        .attr("y", this.yScale(0))
        .attr("height", 0);

    // new bars come in relative to old scale, then animate
    this.xScale = newXScale;
    this.xAxis.scale(this.xScale);

    // delete
    var exitedTransition =
        bars
        .exit()
        .transition()
        .duration(750);

    // TODO: If transition is interrupted, new charts is empty
    exitedTransition
        .attrTween("opacity", function() { return d3.interpolate(1, 0); })
        .remove();
    exitedTransition
        .select("rect")
        .attr("x", function(d) { return self.xScale(d.x); })
        .attr("y", this.yScale(0))
        .attr("height", 0);


    // update (& insert)
    bars
        .select("rect")
        .transition()
        .duration(750)
        .call(positionBar)
        .call(updateBarValue);


    function positionBar(selection) {
        selection
            .attr("x", function(d) { return self.xScale(d.x); })
            .attr("width", function(d) {
                return Math.max(1, (self.xScale(d.x + d.dx) - self.xScale(d.x)) - barMargin);
            })
    }

    function updateBarValue(selection) {
        selection
            .attr("y", function(d) { return self.yScale(d.y); })
            .attr("height", function(d) { return self.yScale(0) - self.yScale(d.y); });
    }


    // Axes
    var xAxisElem = this.svg.select("g.x.axis");
    if (xAxisElem.empty()) {
        xAxisElem =
            this.svg
            .append("g")
            .attr("class", "x axis")
            .attr("transform", "translate(0," + (self.height - self.axisSize[1]) + ")")
            .call(this.xAxis);
    } else {
        xAxisElem
            .transition()
            .duration(750)
            .call(this.xAxis);
    }

    var yAxisElem = this.svg.select("g.y.axis");
    if (yAxisElem.empty()) {
        yAxisElem =
            this.svg
            .append("g")
            .attr("class", "y axis")
            .attr("transform", "translate(" + self.axisSize[0] + ",0)")
            .call(this.yAxis);
    } else {
        yAxisElem
            .transition()
            .duration(750)
            .call(this.yAxis);
    }
}
